@@ -30,6 +30,7 @@
 #include "md_json.h"
 #include "md_jws.h"
 #include "md_log.h"
+#include "md_result.h"
 #include "md_store.h"
 #include "md_util.h"
 #include "md_version.h"
@@ -43,7 +44,6 @@ static apr_status_t acct_make(md_acme_acct_t **pacct, apr_pool_t *p,
     md_acme_acct_t *acct;
     
     acct = apr_pcalloc(p, sizeof(*acct));
-
     acct->ca_url = ca_url;
     if (!contacts || apr_is_empty_array(contacts)) {
         acct->contacts = apr_array_make(p, 5, sizeof(const char *));
@@ -114,7 +114,9 @@ md_json_t *md_acme_acct_to_json(md_acme_acct_t *acct, apr_pool_t *p)
     if (acct->registration) md_json_setj(acct->registration, jacct, MD_KEY_REGISTRATION, NULL);
     if (acct->agreement) md_json_sets(acct->agreement, jacct, MD_KEY_AGREEMENT, NULL);
     if (acct->orders) md_json_sets(acct->orders, jacct, MD_KEY_ORDERS, NULL);
-    
+    if (acct->eab_kid) md_json_sets(acct->eab_kid, jacct, MD_KEY_EAB, MD_KEY_KID, NULL);
+    if (acct->eab_hmac) md_json_sets(acct->eab_hmac, jacct, MD_KEY_EAB, MD_KEY_HMAC, NULL);
+
     return jacct;
 }
 
@@ -129,22 +131,17 @@ apr_status_t md_acme_acct_from_json(md_acme_acct_t **pacct, md_json_t *json, apr
     if (md_json_has_key(json, MD_KEY_STATUS, NULL)) {
         status = acct_st_from_str(md_json_gets(json, MD_KEY_STATUS, NULL));
     }
-    else {
-        /* old accounts only had disabled boolean field */
-        status = md_json_getb(json, MD_KEY_DISABLED, NULL)? 
-            MD_ACME_ACCT_ST_DEACTIVATED : MD_ACME_ACCT_ST_VALID;
-    }
-    
+
     url = md_json_gets(json, MD_KEY_URL, NULL);
     if (!url) {
         md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, p, "account has no url");
-        goto out;
+        goto leave;
     }
 
     ca_url = md_json_gets(json, MD_KEY_CA_URL, NULL);
     if (!ca_url) {
         md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, p, "account has no CA url: %s", url);
-        goto out;
+        goto leave;
     }
     
     contacts = apr_array_make(p, 5, sizeof(const char *));
@@ -155,14 +152,23 @@ apr_status_t md_acme_acct_from_json(md_acme_acct_t **pacct, md_json_t *json, apr
         md_json_getsa(contacts, json, MD_KEY_REGISTRATION, MD_KEY_CONTACT, NULL);
     }
     rv = acct_make(&acct, p, ca_url, contacts);
-    if (APR_SUCCESS == rv) {
-        acct->status = status;
-        acct->url = url;
+    if (APR_SUCCESS != rv) goto leave;
+
+    acct->status = status;
+    acct->url = url;
+    acct->agreement = md_json_gets(json, MD_KEY_AGREEMENT, NULL);
+    if (!acct->agreement) {
+        /* backward compatible check */
         acct->agreement = md_json_gets(json, "terms-of-service", NULL);
-        acct->orders = md_json_gets(json, MD_KEY_ORDERS, NULL);
+    }
+    acct->orders = md_json_gets(json, MD_KEY_ORDERS, NULL);
+    if (md_json_has_key(json, MD_KEY_EAB, MD_KEY_KID, NULL)
+        && md_json_has_key(json, MD_KEY_EAB, MD_KEY_HMAC, NULL)) {
+        acct->eab_kid = md_json_gets(json, MD_KEY_EAB, MD_KEY_KID, NULL);
+        acct->eab_hmac = md_json_gets(json, MD_KEY_EAB, MD_KEY_HMAC, NULL);
     }
 
-out:
+leave:
     *pacct = (APR_SUCCESS == rv)? acct : NULL;
     return rv;
 }
@@ -228,10 +234,36 @@ out:
 /**************************************************************************************************/
 /* Lookup */
 
+int md_acme_acct_matches_url(md_acme_acct_t *acct, const char *url)
+{
+    /* The ACME url must match exactly */
+    if (!url || !acct->ca_url || strcmp(acct->ca_url, url)) return 0;
+    return 1;
+}
+
+int md_acme_acct_matches_md(md_acme_acct_t *acct, const md_t *md)
+{
+    if (!md_acme_acct_matches_url(acct, md->ca_url)) return 0;
+    /* if eab values are not mentioned, we match an account regardless
+     * if it was registered with eab or not */
+    if (!md->ca_eab_kid || !md->ca_eab_hmac) {
+        /* No eab only acceptable when no eab is asked for.
+         * This prevents someone that has no external account binding
+         * to re-use an account from another MDomain that was created
+         * with a binding. */
+        return !acct->eab_kid || !acct->eab_hmac;
+    }
+    /* But of eab is asked for, we need an acct that matches exactly.
+     * When someone configures a new EAB and we need
+     * to created a new account for it. */
+    if (!acct->eab_kid || !acct->eab_hmac) return 0;
+    return !strcmp(acct->eab_kid, md->ca_eab_kid)
+           && !strcmp(acct->eab_hmac, md->ca_eab_hmac);
+}
+
 typedef struct {
     apr_pool_t *p;
-    md_acme_t *acme;
-    int url_match;
+    const md_t *md;
     const char *id;
 } find_ctx;
 
@@ -239,50 +271,46 @@ static int find_acct(void *baton, const char *name, const char *aspect,
                      md_store_vtype_t vtype, void *value, apr_pool_t *ptemp)
 {
     find_ctx *ctx = baton;
-    int disabled;
-    const char *ca_url, *status;
-    
+    md_acme_acct_t *acct;
+    apr_status_t rv;
+
     (void)aspect;
     (void)ptemp;
     md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, ctx->p, "account candidate %s/%s", name, aspect); 
     if (MD_SV_JSON == vtype) {
-        md_json_t *json = value;
-        
-        status = md_json_gets(json, MD_KEY_STATUS, NULL);
-        disabled = md_json_getb(json, MD_KEY_DISABLED, NULL);
-        ca_url = md_json_gets(json, MD_KEY_CA_URL, NULL);
-        
-        if ((!status || !strcmp("valid", status)) && !disabled 
-            && (!ctx->url_match || (ca_url && !strcmp(ctx->acme->url, ca_url)))) {
+        rv = md_acme_acct_from_json(&acct, (md_json_t*)value, ptemp);
+        if (APR_SUCCESS != rv) goto cleanup;
+
+        if (MD_ACME_ACCT_ST_VALID == acct->status
+            && (!ctx->md || md_acme_acct_matches_md(acct, ctx->md))) {
             md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, ctx->p, 
-                          "found account %s for %s: %s, status=%s, disabled=%d, ca-url=%s", 
-                          name, ctx->acme->url, aspect, status, disabled, ca_url);
+                          "found account %s for %s: %s, status=%d",
+                          acct->id, ctx->md->ca_url, aspect, acct->status);
             ctx->id = apr_pstrdup(ctx->p, name);
             return 0;
         }
     }
+cleanup:
     return 1;
 }
 
 static apr_status_t acct_find(const char **pid, md_acme_acct_t **pacct, md_pkey_t **ppkey, 
                               md_store_t *store, md_store_group_t group,
-                              const char *name_pattern, int url_match, 
-                              md_acme_t *acme, apr_pool_t *p)
+                              const char *name_pattern,
+                              const md_t *md, apr_pool_t *p)
 {
     apr_status_t rv;
     find_ctx ctx;
-    
+
+    memset(&ctx, 0, sizeof(ctx));
     ctx.p = p;
-    ctx.acme = acme;
-    ctx.id = NULL;
-    ctx.url_match = url_match;
-    *pid = NULL;
-    
+    ctx.md = md;
+
     rv = md_store_iter(find_acct, &ctx, store, p, group, name_pattern, MD_FN_ACCOUNT, MD_SV_JSON);
     if (ctx.id) {
         *pid = ctx.id;
         rv = md_acme_acct_load(pacct, ppkey, store, group, ctx.id, p);
-        md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "loading account %s", ctx.id);
+        md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "acct_find: got account %s", ctx.id);
     }
     else {
         *pacct = NULL;
@@ -293,19 +321,26 @@ static apr_status_t acct_find(const char **pid, md_acme_acct_t **pacct, md_pkey_
 }
 
 static apr_status_t acct_find_and_verify(md_store_t *store, md_store_group_t group, 
-                                         const char *name_pattern, md_acme_t *acme, apr_pool_t *p)
+                                         const char *name_pattern,
+                                         md_acme_t *acme, const md_t *md,
+                                         apr_pool_t *p)
 {
     md_acme_acct_t *acct;
     md_pkey_t *pkey;
     const char *id;
     apr_status_t rv;
 
-    if (APR_SUCCESS == (rv = acct_find(&id, &acct, &pkey, store, group, name_pattern, 1, acme, p))) {
+    rv = acct_find(&id, &acct, &pkey, store, group, name_pattern, md, p);
+    if (APR_SUCCESS == rv) {
+        md_log_perror(MD_LOG_MARK, MD_LOG_TRACE1, 0, p, "acct_find_and_verify: found %s",
+                      id);
         acme->acct_id = (MD_SG_STAGING == group)? NULL : id;
         acme->acct = acct;
         acme->acct_key = pkey;
-        rv = md_acme_acct_validate(acme, NULL, p);
-    
+        rv = md_acme_acct_validate(acme, (MD_SG_STAGING == group)? NULL : store, p);
+        md_log_perror(MD_LOG_MARK, MD_LOG_TRACE1, rv, p, "acct_find_and_verify: verified %s",
+                      id);
+
         if (APR_SUCCESS != rv) {
             acme->acct_id = NULL;
             acme->acct = NULL;
@@ -320,13 +355,13 @@ static apr_status_t acct_find_and_verify(md_store_t *store, md_store_group_t gro
     return rv;
 }
 
-apr_status_t md_acme_find_acct(md_acme_t *acme, md_store_t *store)
+apr_status_t md_acme_find_acct_for_md(md_acme_t *acme, md_store_t *store, const md_t *md)
 {
     apr_status_t rv;
     
     while (APR_EAGAIN == (rv = acct_find_and_verify(store, MD_SG_ACCOUNTS, 
                                                     mk_acct_pattern(acme->p, acme), 
-                                                    acme, acme->p))) {
+                                                    acme, md, acme->p))) {
         /* nop */
     }
     
@@ -335,61 +370,31 @@ apr_status_t md_acme_find_acct(md_acme_t *acme, md_store_t *store)
          * can already be found in MD_SG_STAGING? */
         md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, acme->p, 
                       "no account found, looking in STAGING");
-        while (APR_EAGAIN == (rv = acct_find_and_verify(store, MD_SG_STAGING, "*", 
-                                                        acme, acme->p))) {
-            /* nop */
+        rv = acct_find_and_verify(store, MD_SG_STAGING, "*", acme, md, acme->p);
+        if (APR_EAGAIN == rv) {
+            rv = APR_ENOENT;
         }
     }
     return rv;
 }
 
-typedef struct {
-    apr_pool_t *p;
-    const char *url;
-    const char *id;
-} load_ctx;
-
-static int id_by_url(void *baton, const char *name, const char *aspect,
-                     md_store_vtype_t vtype, void *value, apr_pool_t *ptemp)
-{
-    load_ctx *ctx = baton;
-    int disabled;
-    const char *acct_url, *status;
-    
-    (void)aspect;
-    (void)ptemp;
-    if (MD_SV_JSON == vtype) {
-        md_json_t *json = value;
-        
-        status = md_json_gets(json, MD_KEY_STATUS, NULL);
-        disabled = md_json_getb(json, MD_KEY_DISABLED, NULL);
-        acct_url = md_json_gets(json, MD_KEY_URL, NULL);
-        
-        if ((!status || !strcmp("valid", status)) && !disabled 
-            && acct_url && !strcmp(ctx->url, acct_url)) {
-            md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, ctx->p, 
-                          "found account %s for url %s: %s, status=%s, disabled=%d", 
-                          name, ctx->url, aspect, status, disabled);
-            ctx->id = apr_pstrdup(ctx->p, name);
-            return 0;
-        }
-    }
-    return 1;
-}
-
-apr_status_t md_acme_acct_id_for_url(const char **pid, md_store_t *store, 
-                                     md_store_group_t group, const char *url, apr_pool_t *p)
+apr_status_t md_acme_acct_id_for_md(const char **pid, md_store_t *store,
+                                    md_store_group_t group, const md_t *md,
+                                    apr_pool_t *p)
 {
     apr_status_t rv;
-    load_ctx ctx;
-    
+    find_ctx ctx;
+
+    memset(&ctx, 0, sizeof(ctx));
     ctx.p = p;
-    ctx.url = url;
-    ctx.id = NULL;
-    
-    rv = md_store_iter(id_by_url, &ctx, store, p, group, "*", MD_FN_ACCOUNT, MD_SV_JSON);
-    *pid = (APR_SUCCESS == rv)? ctx.id : NULL;
-    md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "acct_id_by_url %s -> %s", url, *pid);
+    ctx.md = md;
+
+    rv = md_store_iter(find_acct, &ctx, store, p, group, "*", MD_FN_ACCOUNT, MD_SV_JSON);
+    if (ctx.id) {
+        *pid = ctx.id;
+        rv = APR_SUCCESS;
+    }
+    md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "acct_id_for_md %s -> %s", md->name, *pid);
     return rv;
 }
 
@@ -399,6 +404,8 @@ typedef struct {
     md_acme_t *acme;
     apr_pool_t *p;
     const char *agreement;
+    const char *eab_kid;
+    const char *eab_hmac;
 } acct_ctx_t;
 
 /**************************************************************************************************/
@@ -406,18 +413,8 @@ typedef struct {
 
 static apr_status_t on_init_acct_upd(md_acme_req_t *req, void *baton)
 {
-    md_json_t *jpayload;
-
     (void)baton;
-    jpayload = md_json_create(req->p);
-    switch (MD_ACME_VERSION_MAJOR(req->acme->version)) {
-        case 1:
-            md_json_sets("reg", jpayload, MD_KEY_RESOURCE, NULL);
-            break;
-        default:
-            break;
-    }
-    return md_acme_req_body_init(req, jpayload);
+    return md_acme_req_body_init(req, NULL);
 } 
 
 static apr_status_t acct_upd(md_acme_t *acme, apr_pool_t *p, 
@@ -426,7 +423,12 @@ static apr_status_t acct_upd(md_acme_t *acme, apr_pool_t *p,
     acct_ctx_t *ctx = baton;
     apr_status_t rv = APR_SUCCESS;
     md_acme_acct_t *acct = acme->acct;
-    
+
+    if (md_log_is_level(p, MD_LOG_TRACE2)) {
+        md_log_perror(MD_LOG_MARK, MD_LOG_TRACE2, 0, acme->p, "acct update response: %s",
+                      md_json_writep(body, p, MD_JSON_FMT_COMPACT));
+    }
+
     if (!acct->url) {
         const char *location = apr_table_get(hdrs, "location");
         if (!location) {
@@ -447,6 +449,10 @@ static apr_status_t acct_upd(md_acme_t *acme, apr_pool_t *p,
     if (md_json_has_key(body, MD_KEY_ORDERS, NULL)) {
         acct->orders = md_json_dups(acme->p, body, MD_KEY_ORDERS, NULL);
     }
+    if (ctx->eab_kid && ctx->eab_hmac) {
+        acct->eab_kid = ctx->eab_kid;
+        acct->eab_hmac = ctx->eab_hmac;
+    }
     acct->registration = md_json_clone(ctx->p, body);
     
     md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "updated acct %s", acct->url);
@@ -461,6 +467,7 @@ apr_status_t md_acme_acct_update(md_acme_t *acme)
     if (!acme->acct) {
         return APR_EINVAL;
     }
+    memset(&ctx, 0, sizeof(ctx));
     ctx.acme = acme;
     ctx.p = acme->p;
     return md_acme_POST(acme, acme->acct->url, on_init_acct_upd, acct_upd, NULL, NULL, &ctx);
@@ -471,7 +478,16 @@ apr_status_t md_acme_acct_validate(md_acme_t *acme, md_store_t *store, apr_pool_
     apr_status_t rv;
     
     if (APR_SUCCESS != (rv = md_acme_acct_update(acme))) {
-        if (acme->acct && (APR_ENOENT == rv || APR_EACCES == rv)) {
+        md_log_perror(MD_LOG_MARK, MD_LOG_TRACE1, rv, acme->p,
+                      "acct update failed for %s", acme->acct->url);
+        if (APR_EINVAL == rv && (acme->acct->agreement || !acme->ca_agreement)) {
+            /* Sadly, some proprietary ACME servers choke on empty POSTs
+             * on accounts. Try a faked ToS agreement. */
+            md_log_perror(MD_LOG_MARK, MD_LOG_TRACE1, rv, acme->p,
+                          "trying acct update via ToS agreement");
+            rv = md_acme_agree(acme, p, "accepted");
+        }
+        if (acme->acct && (APR_ENOENT == rv || APR_EACCES == rv || APR_EINVAL == rv)) {
             if (MD_ACME_ACCT_ST_VALID == acme->acct->status) {
                 acme->acct->status = MD_ACME_ACCT_ST_UNKNOWN;
                 if (store) {
@@ -489,34 +505,73 @@ apr_status_t md_acme_acct_validate(md_acme_t *acme, md_store_t *store, apr_pool_
 /**************************************************************************************************/
 /* Register a new account */
 
+static apr_status_t get_eab(md_json_t **peab, md_acme_req_t *req, const char *kid,
+                            const char *hmac64, md_pkey_t *account_key,
+                            const char *url)
+{
+    md_json_t *eab, *prot_fields, *jwk;
+    md_data_t payload, hmac_key;
+    apr_status_t rv;
+
+    prot_fields = md_json_create(req->p);
+    md_json_sets(url, prot_fields, "url", NULL);
+    md_json_sets(kid, prot_fields, "kid", NULL);
+
+    rv = md_jws_get_jwk(&jwk, req->p, account_key);
+    if (APR_SUCCESS != rv) goto cleanup;
+
+    md_data_null(&payload);
+    payload.data = md_json_writep(jwk, req->p, MD_JSON_FMT_COMPACT);
+    if (!payload.data) {
+        rv = APR_EINVAL;
+        goto cleanup;
+    }
+    payload.len = strlen(payload.data);
+
+    md_util_base64url_decode(&hmac_key, hmac64, req->p);
+    if (!hmac_key.len) {
+        rv = APR_EINVAL;
+        md_result_problem_set(req->result, rv, "apache:eab-hmac-invalid",
+                              "external account binding HMAC value is not valid base64", NULL);
+        goto cleanup;
+    }
+
+    rv = md_jws_hmac(&eab, req->p, &payload, prot_fields, &hmac_key);
+    if (APR_SUCCESS != rv) {
+        md_result_problem_set(req->result, rv, "apache:eab-hmac-fail",
+                              "external account binding MAC could not be computed", NULL);
+    }
+
+cleanup:
+    *peab = (APR_SUCCESS == rv)? eab : NULL;
+    return rv;
+}
+
 static apr_status_t on_init_acct_new(md_acme_req_t *req, void *baton)
 {
     acct_ctx_t *ctx = baton;
-    md_json_t *jpayload;
+    md_json_t *jpayload, *jeab;
+    apr_status_t rv;
 
     jpayload = md_json_create(req->p);
-    
-    switch (MD_ACME_VERSION_MAJOR(req->acme->version)) {
-        case 1:
-            md_json_sets("new-reg", jpayload, MD_KEY_RESOURCE, NULL);
-            md_json_setsa(ctx->acme->acct->contacts, jpayload, MD_KEY_CONTACT, NULL);
-            if (ctx->agreement) {
-                md_json_sets(ctx->agreement, jpayload, MD_KEY_AGREEMENT, NULL);
-            }
-            break;
-        default:
-            md_json_setsa(ctx->acme->acct->contacts, jpayload, MD_KEY_CONTACT, NULL);
-            if (ctx->agreement) {
-                md_json_setb(1, jpayload, "termsOfServiceAgreed", NULL);
-            }
-        break;
+    md_json_setsa(ctx->acme->acct->contacts, jpayload, MD_KEY_CONTACT, NULL);
+    if (ctx->agreement) {
+        md_json_setb(1, jpayload, "termsOfServiceAgreed", NULL);
     }
-    
-    return md_acme_req_body_init(req, jpayload);
+    if (ctx->eab_kid && ctx->eab_hmac) {
+        rv = get_eab(&jeab, req, ctx->eab_kid, ctx->eab_hmac,
+                     req->acme->acct_key, req->url);
+        if (APR_SUCCESS != rv) goto cleanup;
+        md_json_setj(jeab, jpayload, "externalAccountBinding", NULL);
+    }
+    rv = md_acme_req_body_init(req, jpayload);
+
+cleanup:
+    return rv;
 } 
 
-apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store, apr_pool_t *p, 
-                                   apr_array_header_t *contacts, const char *agreement)
+apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store,
+                                   const md_t *md, apr_pool_t *p)
 {
     apr_status_t rv;
     md_pkey_t *pkey;
@@ -526,15 +581,17 @@ apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store, apr_pool_
     acct_ctx_t ctx;
     
     md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, p, "create new account");
-    
+
+    memset(&ctx, 0, sizeof(ctx));
     ctx.acme = acme;
     ctx.p = p;
     /* The agreement URL is submitted when the ACME server announces Terms-of-Service
      * in its directory meta data. The magic value "accepted" will always use the
      * advertised URL. */
     ctx.agreement = NULL;
-    if (acme->ca_agreement && agreement) {
-        ctx.agreement = !strcmp("accepted", agreement)? acme->ca_agreement : agreement;
+    if (acme->ca_agreement && md->ca_agreement) {
+        ctx.agreement = !strcmp("accepted", md->ca_agreement)?
+            acme->ca_agreement : md->ca_agreement;
     }
     
     if (ctx.agreement) {
@@ -544,9 +601,11 @@ apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store, apr_pool_
             goto out;
         }
     }
+    ctx.eab_kid = md->ca_eab_kid;
+    ctx.eab_hmac = md->ca_eab_hmac;
     
-    for (i = 0; i < contacts->nelts; ++i) {
-        uri = APR_ARRAY_IDX(contacts, i, const char *);
+    for (i = 0; i < md->contacts->nelts; ++i) {
+        uri = APR_ARRAY_IDX(md->contacts, i, const char *);
         if (APR_SUCCESS != (rv = md_util_abs_uri_check(acme->p, uri, &err))) {
             md_log_perror(MD_LOG_MARK, MD_LOG_ERR, 0, p, 
                           "invalid contact uri (%s): %s", err, uri);
@@ -563,19 +622,19 @@ apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store, apr_pool_
      */
     if (!acme->acct_key) {
         find_ctx fctx;
-    
+
+        memset(&fctx, 0, sizeof(fctx));
         fctx.p = p;
-        fctx.acme = acme;
-        fctx.id = NULL;
-        fctx.url_match = 0;
-        
+        fctx.md = md;
+
         md_store_iter(find_acct, &fctx, store, p, MD_SG_ACCOUNTS, 
                       mk_acct_pattern(p, acme), MD_FN_ACCOUNT, MD_SV_JSON);
         if (fctx.id) {
-            rv = md_store_load(store, MD_SG_ACCOUNTS, fctx.id, MD_FN_ACCT_KEY, MD_SV_PKEY, 
+            rv = md_store_load(store, MD_SG_ACCOUNTS, fctx.id, MD_FN_ACCT_KEY, MD_SV_PKEY,
                                (void**)&acme->acct_key, p);
             if (APR_SUCCESS == rv) {
-                md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "reusing key from account %s", fctx.id);
+                md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p,
+                              "reusing key from account %s", fctx.id);
             }
             else {
                 acme->acct_key = NULL;
@@ -593,7 +652,7 @@ apr_status_t md_acme_acct_register(md_acme_t *acme, md_store_t *store, apr_pool_
         md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, rv, p, "created new account key");
     }
     
-    if (APR_SUCCESS != (rv = acct_make(&acme->acct,  p, acme->url, contacts))) goto out;
+    if (APR_SUCCESS != (rv = acct_make(&acme->acct,  p, acme->url, md->contacts))) goto out;
     rv = md_acme_POST_new_account(acme,  on_init_acct_new, acct_upd, NULL, NULL, &ctx);
     if (APR_SUCCESS == rv) {
         md_log_perror(MD_LOG_MARK, MD_LOG_INFO, 0, p, 
@@ -616,15 +675,7 @@ static apr_status_t on_init_acct_del(md_acme_req_t *req, void *baton)
 
     (void)baton;
     jpayload = md_json_create(req->p);
-    switch (MD_ACME_VERSION_MAJOR(req->acme->version)) {
-        case 1:
-            md_json_sets("reg", jpayload, MD_KEY_RESOURCE, NULL);
-            md_json_setb(1, jpayload, "delete", NULL);
-            break;
-        default:
-            md_json_sets("deactivated", jpayload, MD_KEY_STATUS, NULL);
-            break;
-    }
+    md_json_sets("deactivated", jpayload, MD_KEY_STATUS, NULL);
     return md_acme_req_body_init(req, jpayload);
 } 
 
@@ -639,6 +690,7 @@ apr_status_t md_acme_acct_deactivate(md_acme_t *acme, apr_pool_t *p)
     }
     md_log_perror(MD_LOG_MARK, MD_LOG_DEBUG, 0, acme->p, "delete account %s from %s", 
                   acct->url, acct->ca_url);
+    memset(&ctx, 0, sizeof(ctx));
     ctx.acme = acme;
     ctx.p = p;
     return md_acme_POST(acme, acct->url, on_init_acct_del, acct_upd, NULL, NULL, &ctx);
@@ -653,16 +705,8 @@ static apr_status_t on_init_agree_tos(md_acme_req_t *req, void *baton)
     md_json_t *jpayload;
 
     jpayload = md_json_create(req->p);
-    switch (MD_ACME_VERSION_MAJOR(req->acme->version)) {
-        case 1:
-            md_json_sets("reg", jpayload, MD_KEY_RESOURCE, NULL);
-            md_json_sets(ctx->acme->acct->agreement, jpayload, MD_KEY_AGREEMENT, NULL);
-            break;
-        default:
-            if (ctx->acme->acct->agreement) {
-                md_json_setb(1, jpayload, "termsOfServiceAgreed", NULL);
-            }
-            break;
+    if (ctx->acme->acct->agreement) {
+        md_json_setb(1, jpayload, "termsOfServiceAgreed", NULL);
     }
     return md_acme_req_body_init(req, jpayload);
 } 
@@ -676,6 +720,7 @@ apr_status_t md_acme_agree(md_acme_t *acme, apr_pool_t *p, const char *agreement
         acme->acct->agreement = acme->ca_agreement;
     }
     
+    memset(&ctx, 0, sizeof(ctx));
     ctx.acme = acme;
     ctx.p = p;
     return md_acme_POST(acme, acme->acct->url, on_init_agree_tos, acct_upd, NULL, NULL, &ctx);
