@@ -32,6 +32,7 @@
 #include "md_curl.h"
 #include "md_crypt.h"
 #include "md_http.h"
+#include "md_ocsp.h"
 #include "md_json.h"
 #include "md_status.h"
 #include "md_store.h"
@@ -57,9 +58,12 @@
 
 int md_http_cert_status(request_rec *r)
 {
-    md_json_t *resp, *j, *mdj, *certj;
+    int i;
+    md_json_t *resp, *mdj, *cj;
     const md_srv_conf_t *sc;
     const md_t *md;
+    md_pkey_spec_t *spec;
+    const char *keyname;
     apr_bucket_brigade *bb;
     apr_status_t rv;
     
@@ -84,8 +88,9 @@ int md_http_cert_status(request_rec *r)
     ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r,
                   "requesting status for MD: %s", md->name);
 
-    if (APR_SUCCESS != (rv = md_status_get_md_json(&mdj, md, sc->mc->reg, r->pool))) {
-        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(10175)
+    rv = md_status_get_md_json(&mdj, md, sc->mc->reg, sc->mc->ocsp, r->pool);
+    if (APR_SUCCESS != rv) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r, APLOGNO(10204)
                       "loading md status for %s", md->name);
         return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -94,32 +99,40 @@ int md_http_cert_status(request_rec *r)
                   "status for MD: %s is %s", md->name, md_json_writep(mdj, r->pool, MD_JSON_FMT_INDENT));
 
     resp = md_json_create(r->pool);
-    
-    if (md_json_has_key(mdj, MD_KEY_CERT, MD_KEY_VALID_UNTIL, NULL)) {
-        md_json_sets(md_json_gets(mdj, MD_KEY_CERT, MD_KEY_VALID_UNTIL, NULL), 
-                     resp, MD_KEY_VALID_UNTIL, NULL);
-    }
-    if (md_json_has_key(mdj, MD_KEY_CERT, MD_KEY_VALID_FROM, NULL)) {
-        md_json_sets(md_json_gets(mdj, MD_KEY_CERT, MD_KEY_VALID_FROM, NULL), 
-                     resp, MD_KEY_VALID_FROM, NULL);
-    }
-    if (md_json_has_key(mdj, MD_KEY_CERT, MD_KEY_SERIAL, NULL)) {
-        md_json_sets(md_json_gets(mdj, MD_KEY_CERT, MD_KEY_SERIAL, NULL), 
-                     resp, MD_KEY_SERIAL, NULL);
-    }
-    if (md_json_has_key(mdj, MD_KEY_CERT, MD_KEY_SHA256_FINGERPRINT, NULL)) {
-        md_json_sets(md_json_gets(mdj, MD_KEY_CERT, MD_KEY_SHA256_FINGERPRINT, NULL), 
-                     resp, MD_KEY_SHA256_FINGERPRINT, NULL);
+
+    if (md_json_has_key(mdj, MD_KEY_CERT, MD_KEY_VALID, NULL)) {
+         md_json_setj(md_json_getj(mdj, MD_KEY_CERT, MD_KEY_VALID, NULL), resp, MD_KEY_VALID, NULL);
+     }
+
+    for (i = 0; i < md_cert_count(md); ++i) {
+        spec = md_pkeys_spec_get(md->pks, i);
+        keyname = md_pkey_spec_name(spec);
+        cj = md_json_create(r->pool);
+
+        if (md_json_has_key(mdj, MD_KEY_CERT, keyname, MD_KEY_VALID, NULL)) {
+            md_json_setj(md_json_getj(mdj, MD_KEY_CERT, keyname, MD_KEY_VALID, NULL),
+                         cj, MD_KEY_VALID, NULL);
+        }
+
+        if (md_json_has_key(mdj, MD_KEY_CERT, keyname, MD_KEY_SERIAL, NULL)) {
+            md_json_sets(md_json_gets(mdj, MD_KEY_CERT, keyname, MD_KEY_SERIAL, NULL),
+                         cj, MD_KEY_SERIAL, NULL);
+        }
+        if (md_json_has_key(mdj, MD_KEY_CERT, keyname, MD_KEY_SHA256_FINGERPRINT, NULL)) {
+            md_json_sets(md_json_gets(mdj, MD_KEY_CERT, keyname, MD_KEY_SHA256_FINGERPRINT, NULL),
+                         cj, MD_KEY_SHA256_FINGERPRINT, NULL);
+        }
+        md_json_setj(cj, resp, keyname, NULL );
     }
     
     if (md_json_has_key(mdj, MD_KEY_RENEWAL, NULL)) {
-        /* copy over the information we want to make public about this:
-         *  - when not finished, add an empty object to indicate something is going on
-         *  - when a certificate is staged, add the information from that */
-        certj = md_json_getj(mdj, MD_KEY_RENEWAL, MD_KEY_CERT, NULL);
-        j = certj? certj : md_json_create(r->pool);; 
-        md_json_setj(j, resp, MD_KEY_RENEWAL, NULL);
-    }
+           /* copy over the information we want to make public about this:
+            *  - when not finished, add an empty object to indicate something is going on
+            *  - when a certificate is staged, add the information from that */
+           cj = md_json_getj(mdj, MD_KEY_RENEWAL, MD_KEY_CERT, NULL);
+           cj = cj? cj : md_json_create(r->pool);; 
+           md_json_setj(cj, resp, MD_KEY_RENEWAL, MD_KEY_CERT, NULL);
+     }
     
     ap_log_rerror(APLOG_MARK, APLOG_TRACE2, 0, r, "md[%s]: sending status", md->name);
     apr_table_set(r->headers_out, "Content-Type", "application/json"); 
@@ -156,11 +169,18 @@ struct status_info {
 static void si_val_status(status_ctx *ctx, md_json_t *mdj, const status_info *info)
 {
     const char *s = "unknown";
+    apr_time_t until;
     (void)info;
-    switch (md_json_getl(mdj, MD_KEY_STATE, NULL)) {
-        case MD_S_INCOMPLETE: s = "incomplete"; break;
+    switch (md_json_getl(mdj, info->key, NULL)) {
+        case MD_S_INCOMPLETE:
+            s = md_json_gets(mdj, MD_KEY_STATE_DESCR, NULL);
+            s = s? apr_psprintf(ctx->p, "incomplete: %s", s) : "incomplete";
+            break;
         case MD_S_EXPIRED_DEPRECATED:
-        case MD_S_COMPLETE: s = "ok"; break;
+        case MD_S_COMPLETE:
+            until = md_json_get_time(mdj, MD_KEY_CERT, MD_KEY_VALID, MD_KEY_UNTIL, NULL);
+            s = (!until || until > apr_time_now())? "good" : "expired"; 
+            break;
         case MD_S_ERROR: s = "error"; break;
         case MD_S_MISSING_INFORMATION: s = "missing information"; break;
         default: break;
@@ -168,19 +188,19 @@ static void si_val_status(status_ctx *ctx, md_json_t *mdj, const status_info *in
     apr_brigade_puts(ctx->bb, NULL, NULL, s);
 }
 
-static void si_val_renew_mode(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+static void si_val_url(status_ctx *ctx, md_json_t *mdj, const status_info *info)
 {
-    const char *s;
-    switch (md_json_getl(mdj, info->key, NULL)) {
-        case MD_RENEW_MANUAL: s = "manual"; break;
-        case MD_RENEW_ALWAYS: s = "always"; break;
-        default: s = "auto"; break;
-    }
-    apr_brigade_puts(ctx->bb, NULL, NULL, s);
+    const char *url, *s;
+
+    s = url = md_json_gets(mdj, info->key, NULL);
+    if (!url) return;
+    s = md_get_ca_name_from_url(ctx->p, url);
+    apr_brigade_printf(ctx->bb, NULL, NULL, "<a href='%s'>%s</a>",
+                       ap_escape_html2(ctx->p, url, 1), 
+                       ap_escape_html2(ctx->p, s, 1));
 }
 
-
-static void si_val_date(status_ctx *ctx, apr_time_t timestamp)
+static void print_date(apr_bucket_brigade *bb, apr_time_t timestamp, const char *title)
 {
     if (timestamp > 0) {
         char ts[128];
@@ -189,189 +209,230 @@ static void si_val_date(status_ctx *ctx, apr_time_t timestamp)
         apr_size_t len;
         
         apr_time_exp_gmt(&texp, timestamp);
-        apr_strftime(ts, &len, sizeof(ts)-1, "%Y-%m-%dT%H:%M:%SZ", &texp);
+        apr_strftime(ts, &len, sizeof(ts2)-1, "%Y-%m-%d", &texp);
         ts[len] = '\0';
+        if (!title) {
+            apr_strftime(ts2, &len, sizeof(ts)-1, "%Y-%m-%dT%H:%M:%SZ", &texp);
+            ts2[len] = '\0';
+            title = ts2;
+        }
+        apr_brigade_printf(bb, NULL, NULL, 
+                           "<span title='%s' style='white-space: nowrap;'>%s</span>", 
+                           ap_escape_html2(bb->p, title, 1), ts);
+    }
+}
+
+static void print_time(apr_bucket_brigade *bb, const char *label, apr_time_t t)
+{
+    apr_time_t now;
+    const char *pre, *post, *sep;
+    char ts[APR_RFC822_DATE_LEN];
+    char ts2[128];
+    apr_time_exp_t texp;
+    apr_size_t len;
+    apr_interval_time_t delta;
+    
+    if (t == 0) {
+        /* timestamp is 0, we use that for "not set" */
+        return;
+    }
+    apr_time_exp_gmt(&texp, t);
+    now = apr_time_now();
+    pre = post = "";
+    sep = (label && strlen(label))? " " : "";
+    delta = 0;
+    apr_rfc822_date(ts, t);
+    if (t > now) {
+        delta = t - now;
+        pre = "in ";
+    }
+    else {
+        delta = now - t;
+        post = " ago";
+    }
+    if (delta >= (4 * apr_time_from_sec(MD_SECS_PER_DAY))) {
         apr_strftime(ts2, &len, sizeof(ts2)-1, "%Y-%m-%d", &texp);
         ts2[len] = '\0';
-        apr_brigade_printf(ctx->bb, NULL, NULL, 
-                           "<span title='%s' style='white-space: nowrap;'>%s</span>", 
-                           ts, ts2);
+        apr_brigade_printf(bb, NULL, NULL, "%s%s<span title='%s' "
+                           "style='white-space: nowrap;'>%s</span>", 
+                           label, sep, ts, ts2); 
     }
     else {
-        apr_brigade_puts(ctx->bb, NULL, NULL, "-");
+        apr_brigade_printf(bb, NULL, NULL, "%s%s<span title='%s'>%s%s%s</span>", 
+                           label, sep, ts, pre, md_duration_roughly(bb->p, delta), post); 
     }
 }
 
-static void si_val_time(status_ctx *ctx, apr_time_t timestamp)
+static void si_val_valid_time(status_ctx *ctx, md_json_t *mdj, const status_info *info)
 {
-    if (timestamp > 0) {
-        char ts[128];
-        char ts2[128];
-        apr_time_exp_t texp;
-        apr_size_t len;
-        
-        apr_time_exp_gmt(&texp, timestamp);
-        apr_strftime(ts, &len, sizeof(ts)-1, "%Y-%m-%dT%H:%M:%SZ", &texp);
-        ts[len] = '\0';
-        apr_strftime(ts2, &len, sizeof(ts2)-1, "%H:%M:%SZ", &texp);
-        ts2[len] = '\0';
-        apr_brigade_printf(ctx->bb, NULL, NULL, 
-                           "<span title='%s' style='white-space: nowrap;'>%s</span>", 
-                           ts, ts2);
-    }
-    else {
-        apr_brigade_puts(ctx->bb, NULL, NULL, "-");
-    }
-}
-
-static void si_val_expires(status_ctx *ctx, md_json_t *mdj, const status_info *info)
-{
-    const char *s;
-    apr_time_t t;
+    const char *sfrom, *suntil, *sep, *title;
+    apr_time_t from, until;
     
-    (void)info;
-    s = md_json_dups(ctx->p, mdj, MD_KEY_CERT, MD_KEY_VALID_UNTIL, NULL);
-    if (s) {
-        t = apr_date_parse_rfc(s);
-        si_val_date(ctx, t);
-    }
-}
-
-static void si_val_valid_from(status_ctx *ctx, md_json_t *mdj, const status_info *info)
-{
-    const char *s;
-    apr_time_t t;
+    sep = NULL;
+    sfrom = md_json_gets(mdj, info->key, MD_KEY_FROM, NULL);
+    from = sfrom? apr_date_parse_rfc(sfrom) : 0;
+    suntil = md_json_gets(mdj, info->key, MD_KEY_UNTIL, NULL);
+    until = suntil?apr_date_parse_rfc(suntil) : 0;
     
-    (void)info;
-    s = md_json_dups(ctx->p, mdj, MD_KEY_CERT, MD_KEY_VALID_FROM, NULL);
-    if (s) {
-        t = apr_date_parse_rfc(s);
-        si_val_date(ctx, t);
+    if (from > apr_time_now()) {
+        apr_brigade_puts(ctx->bb, NULL, NULL, "from ");
+        print_date(ctx->bb, from, sfrom);
+        sep = " ";
+    }
+    if (until) {
+        if (sep) apr_brigade_puts(ctx->bb, NULL, NULL, sep);
+        apr_brigade_puts(ctx->bb, NULL, NULL, "until ");
+        title = sfrom? apr_psprintf(ctx->p, "%s - %s", sfrom, suntil) : suntil;
+        print_date(ctx->bb, until, title);
     }
 }
-    
-static void si_val_props(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+
+static void si_add_header(status_ctx *ctx, const status_info *info)
 {
-    const char *s, *url;
-    md_pkey_type_t ptype;
-    int i = 0;
-    (void)info;
-
-    if (md_json_getb(mdj, MD_KEY_MUST_STAPLE, NULL)) {
-        ++i;
-        apr_brigade_puts(ctx->bb, NULL, NULL, "must-staple");
-    }
-    s = md_json_gets(mdj, MD_KEY_RENEW_WINDOW, NULL);
-    if (s) {
-        if (i++) apr_brigade_puts(ctx->bb, NULL, NULL, " \n"); 
-        apr_brigade_printf(ctx->bb, NULL, NULL, "renew-at[%s]", s);
-    }
-    url = s = md_json_gets(mdj, MD_KEY_CA, MD_KEY_URL, NULL);
-    if (s) {
-        if (i++) apr_brigade_puts(ctx->bb, NULL, NULL, " \n"); 
-        if (!strcmp(LE_ACMEv2_PROD, s)) s = "letsencrypt(v2)";
-        else if (!strcmp(LE_ACMEv1_PROD, s)) s = "letsencrypt(v1)";
-        else if (!strcmp(LE_ACMEv2_STAGING, s)) s = "letsencrypt(Testv2)";
-        else if (!strcmp(LE_ACMEv1_STAGING, s)) s = "letsencrypt(Testv1)";
-        
-        apr_brigade_printf(ctx->bb, NULL, NULL, "ca=[<a href=\"%s\">%s</a>]", url, s);
-    }
-    if (md_json_has_key(mdj, MD_KEY_CONTACTS, NULL)) {
-        if (i++) apr_brigade_puts(ctx->bb, NULL, NULL, " \n"); 
-        apr_brigade_puts(ctx->bb, NULL, NULL, "contacts=[");
-        add_json_val(ctx, md_json_getj(mdj, MD_KEY_CONTACTS, NULL));
-        apr_brigade_puts(ctx->bb, NULL, NULL, "]");
-    }
-    ptype = md_json_has_key(mdj, MD_KEY_PKEY, MD_KEY_TYPE, NULL)?
-            (unsigned)md_json_getl(mdj, MD_KEY_PKEY, MD_KEY_TYPE, NULL) : MD_PKEY_TYPE_DEFAULT; 
-    switch (ptype) {
-        case MD_PKEY_TYPE_RSA:
-            if (i++) apr_brigade_puts(ctx->bb, NULL, NULL, " \n"); 
-            apr_brigade_printf(ctx->bb, NULL, NULL, "key[RSA(%u)]", 
-                (unsigned)md_json_getl(mdj, MD_KEY_PKEY, MD_PKEY_RSA_BITS_MIN, NULL));
-        default:
-            break;
-    }
+    const char *html = ap_escape_html2(ctx->p, info->label, 1);
+    apr_brigade_printf(ctx->bb, NULL, NULL, "<th class=\"%s\">%s</th>", html, html);
 }
 
-static void si_val_renewal(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+static void si_val_cert_valid_time(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    md_json_t *jcert;
+    status_info sub = *info;
+    
+    sub.key = MD_KEY_VALID;
+    jcert = md_json_getj(mdj, info->key, NULL);
+    if (jcert) si_val_valid_time(ctx, jcert, &sub);
+}
+
+static void si_val_ca_url(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    md_json_t *jcert;
+    status_info sub = *info;
+    
+    sub.key = MD_KEY_URL;
+    jcert = md_json_getj(mdj, info->key, NULL);
+    if (jcert) si_val_url(ctx, jcert, &sub);
+}
+
+static int count_certs(void *baton, const char *key, md_json_t *json)
+{
+    int *pcount = baton;
+
+    (void)json;
+    if (strcmp(key, MD_KEY_VALID)) {
+        *pcount += 1;
+    }
+    return 1;
+}
+
+static void print_job_summary(apr_bucket_brigade *bb, md_json_t *mdj, const char *key, 
+                              const char *separator)
 {
     char buffer[HUGE_STRING_LEN];
     apr_status_t rv;
-    int finished, errors;
+    int finished, errors, cert_count;
     apr_time_t t;
-    const char *s;
+    const char *s, *line;
     
-    (void)info;
-    if (!md_json_has_key(mdj, MD_KEY_RENEWAL, NULL)) {
+    if (!md_json_has_key(mdj, key, NULL)) {
         return;
     }
     
-    finished = (int)md_json_getl(mdj, MD_KEY_RENEWAL, MD_KEY_FINISHED, NULL);
-    errors = (int)md_json_getl(mdj, MD_KEY_RENEWAL, MD_KEY_ERRORS, NULL);
-    rv = (apr_status_t)md_json_getl(mdj, MD_KEY_RENEWAL, MD_KEY_LAST, MD_KEY_STATUS, NULL);
+    finished = md_json_getb(mdj, key, MD_KEY_FINISHED, NULL);
+    errors = (int)md_json_getl(mdj, key, MD_KEY_ERRORS, NULL);
+    rv = (apr_status_t)md_json_getl(mdj, key, MD_KEY_LAST, MD_KEY_STATUS, NULL);
     
+    line = separator? separator : "";
+
     if (rv != APR_SUCCESS) {
-        s = md_json_gets(mdj, MD_KEY_RENEWAL, MD_KEY_LAST, MD_KEY_PROBLEM, NULL);
-        apr_brigade_printf(ctx->bb, NULL, NULL, "Error[%s]: %s", 
+        s = md_json_gets(mdj, key, MD_KEY_LAST, MD_KEY_PROBLEM, NULL);
+        line = apr_psprintf(bb->p, "%s Error[%s]: %s", line, 
                            apr_strerror(rv, buffer, sizeof(buffer)), s? s : "");
     }
     
     if (finished) {
-        apr_brigade_puts(ctx->bb, NULL, NULL, "Finished");
-        if (md_json_has_key(mdj, MD_KEY_RENEWAL, MD_KEY_VALID_FROM, NULL)) {
-            s = md_json_gets(mdj,  MD_KEY_RENEWAL, MD_KEY_VALID_FROM, NULL);
-            t = apr_date_parse_rfc(s);
-            apr_brigade_puts(ctx->bb, NULL, NULL, (apr_time_now() >= t)?
-                             ", valid since: " : ", activate at: ");
-            si_val_time(ctx, t);
+        cert_count = 0;
+        md_json_iterkey(count_certs, &cert_count, mdj, key, MD_KEY_CERT, NULL);
+        if (cert_count > 0) {
+            line =apr_psprintf(bb->p, "%s  finished, %d new certificate%s staged.",
+                               line, cert_count, cert_count > 1? "s" : "");
         }
-        apr_brigade_puts(ctx->bb, NULL, NULL, ".");
-    } 
-    
-    s = md_json_gets(mdj, MD_KEY_RENEWAL, MD_KEY_LAST, MD_KEY_DETAIL, NULL);
-    if (s) apr_brigade_puts(ctx->bb, NULL, NULL, s);
+        else {
+            line = apr_psprintf(bb->p, "%s  finished successfully.", line);
+        }
+    }
+    else {
+        s = md_json_gets(mdj, key, MD_KEY_LAST, MD_KEY_DETAIL, NULL);
+        if (s) line = apr_psprintf(bb->p, "%s %s", line, s);
+    }
     
     errors = (int)md_json_getl(mdj, MD_KEY_ERRORS, NULL);
     if (errors > 0) {
-        apr_brigade_printf(ctx->bb, NULL, NULL, ", Had %d errors.", errors);
+        line = apr_psprintf(bb->p, "%s (%d retr%s) ", line, 
+            errors, (errors > 1)? "y" : "ies");
     } 
     
-    s = md_json_gets(mdj,  MD_KEY_RENEWAL, MD_KEY_NEXT_RUN, NULL);
-    if (s) {
-        t = apr_date_parse_rfc(s);
-        apr_brigade_puts(ctx->bb, NULL, NULL, "Next attempt: ");
-        si_val_time(ctx, t);
-        apr_brigade_puts(ctx->bb, NULL, NULL, ".");
+    apr_brigade_puts(bb, NULL, NULL, line);
+
+    t = md_json_get_time(mdj, key, MD_KEY_NEXT_RUN, NULL);
+    if (t > apr_time_now() && !finished) {
+        print_time(bb, "\nNext run", t);
     }
+    else if (!strlen(line)) {
+        apr_brigade_puts(bb, NULL, NULL, "\nOngoing...");
+    }
+}
+
+static void si_val_activity(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    apr_time_t t;
+    
+    (void)info;
+    if (md_json_has_key(mdj, MD_KEY_RENEWAL, NULL)) {
+        print_job_summary(ctx->bb, mdj, MD_KEY_RENEWAL, NULL);
+        return;
+    }
+    
+    t = md_json_get_time(mdj, MD_KEY_RENEW_AT, NULL);
+    if (t > apr_time_now()) {
+        print_time(ctx->bb, "Renew", t);
+    }
+    else if (t) {
+        apr_brigade_puts(ctx->bb, NULL, NULL, "Pending");
+    }
+    else if (MD_RENEW_MANUAL == md_json_getl(mdj, MD_KEY_RENEW_MODE, NULL)) {
+        apr_brigade_puts(ctx->bb, NULL, NULL, "Manual renew");
+    }
+}
+
+static int cert_check_iter(void *baton, const char *key, md_json_t *json)
+{
+    status_ctx *ctx = baton;
+    const char *fingerprint;
+    
+    fingerprint = md_json_gets(json, MD_KEY_SHA256_FINGERPRINT, NULL);
+    if (fingerprint) {
+        apr_brigade_printf(ctx->bb, NULL, NULL, 
+                           "<a href=\"%s%s\">%s[%s]</a><br>", 
+                           ctx->mc->cert_check_url, fingerprint, 
+                           ctx->mc->cert_check_name, key);
+    }
+    return 1;
 }
 
 static void si_val_remote_check(status_ctx *ctx, md_json_t *mdj, const status_info *info)
 {
-    const char *fingerprint;
-    
     (void)info;
-    fingerprint = md_json_gets(mdj, MD_KEY_CERT, MD_KEY_SHA256_FINGERPRINT, NULL);
-    if (fingerprint) {
-        apr_brigade_printf(ctx->bb, NULL, NULL, 
-                           "<a href=\"https://censys.io/certificates/%s\">censys.io</a> ", 
-                           fingerprint);
-        apr_brigade_printf(ctx->bb, NULL, NULL, 
-                           "<a href=\"https://crt.sh?q=%s\">crt.sh</a> ", fingerprint);
+    if (ctx->mc->cert_check_name && ctx->mc->cert_check_url) {
+        md_json_iterkey(cert_check_iter, ctx, mdj, MD_KEY_CERT, NULL);
     }
 }
 
-const status_info status_infos[] = {
-    { "Name", MD_KEY_NAME, NULL },
-    { "Domains", MD_KEY_DOMAINS, NULL },
-    { "Status", MD_KEY_STATUS, si_val_status },
-    { "Valid", MD_KEY_VALID_FROM, si_val_valid_from },
-    { "Expires", MD_KEY_VALID_UNTIL, si_val_expires },
-    { "Renew", MD_KEY_RENEW_MODE, si_val_renew_mode },
-    { "Check@", MD_KEY_SHA256_FINGERPRINT, si_val_remote_check },
-    { "Configuration", MD_KEY_MUST_STAPLE, si_val_props },
-    { "Renewal",  MD_KEY_NOTIFIED, si_val_renewal },
-};
+static void si_val_stapling(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    (void)info;
+    if (!md_json_getb(mdj, MD_KEY_STAPLING, NULL)) return;
+    apr_brigade_puts(ctx->bb, NULL, NULL, "on");
+}
 
 static int json_iter_val(void *data, size_t index, md_json_t *json)
 {
@@ -396,6 +457,16 @@ static void add_json_val(status_ctx *ctx, md_json_t *j)
     else if (md_json_is(MD_JSON_TYPE_OBJECT, j, NULL)) {
         md_json_writeb(j, MD_JSON_FMT_COMPACT, ctx->bb);
     }
+    else if (md_json_is(MD_JSON_TYPE_BOOL, j, NULL)) {
+        apr_brigade_puts(ctx->bb, NULL, NULL, md_json_getb(j, NULL)? "on" : "off");
+    }
+}
+
+static void si_val_names(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    apr_brigade_puts(ctx->bb, NULL, NULL, "<div style=\"max-width:400px;\">");
+    add_json_val(ctx, md_json_getj(mdj, info->key, NULL));
+    apr_brigade_puts(ctx->bb, NULL, NULL, "</div>");
 }
 
 static void add_status_cell(status_ctx *ctx, md_json_t *mdj, const status_info *info)
@@ -408,6 +479,17 @@ static void add_status_cell(status_ctx *ctx, md_json_t *mdj, const status_info *
     }
 }
 
+static const status_info status_infos[] = {
+    { "Domain", MD_KEY_NAME, NULL },
+    { "Names", MD_KEY_DOMAINS, si_val_names },
+    { "Status", MD_KEY_STATE, si_val_status },
+    { "Valid", MD_KEY_CERT, si_val_cert_valid_time },
+    { "CA", MD_KEY_CA, si_val_ca_url },
+    { "Stapling", MD_KEY_STAPLING, si_val_stapling },
+    { "Check@", MD_KEY_SHA256_FINGERPRINT, si_val_remote_check },
+    { "Activity",  MD_KEY_NOTIFIED, si_val_activity },
+};
+
 static int add_md_row(void *baton, apr_size_t index, md_json_t *mdj)
 {
     status_ctx *ctx = baton;
@@ -428,7 +510,7 @@ static int md_name_cmp(const void *v1, const void *v2)
     return strcmp((*(const md_t**)v1)->name, (*(const md_t**)v2)->name);
 }
 
-int md_status_hook(request_rec *r, int flags)
+int md_domains_status_hook(request_rec *r, int flags)
 {
     const md_srv_conf_t *sc;
     const md_mod_conf_t *mc;
@@ -437,6 +519,7 @@ int md_status_hook(request_rec *r, int flags)
     apr_array_header_t *mds;
     md_json_t *jstatus, *jstock;
     
+    ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "server-status for managed domains, start");
     sc = ap_get_module_config(r->server->module_config, &md_module);
     if (!sc) return DECLINED;
     mc = sc->mc;
@@ -452,9 +535,11 @@ int md_status_hook(request_rec *r, int flags)
     qsort(mds->elts, (size_t)mds->nelts, sizeof(md_t *), md_name_cmp);
 
     if (!html) {
-        apr_brigade_puts(ctx.bb, NULL, NULL, "ManagedDomains: ");
+        ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "no-html summary");
+        apr_brigade_puts(ctx.bb, NULL, NULL, "Managed Certificates: ");
         if (mc->mds->nelts > 0) {
             md_status_take_stock(&jstock, mds, mc->reg, r->pool);
+            ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "got JSON summary");
             apr_brigade_printf(ctx.bb, NULL, NULL, "total=%d, ok=%d renew=%d errored=%d ready=%d",
                                 (int)md_json_getl(jstock, MD_KEY_TOTAL, NULL), 
                                 (int)md_json_getl(jstock, MD_KEY_COMPLETE, NULL), 
@@ -468,13 +553,13 @@ int md_status_hook(request_rec *r, int flags)
         apr_brigade_puts(ctx.bb, NULL, NULL, "\n"); 
     }
     else if (mc->mds->nelts > 0) {
-        md_status_get_json(&jstatus, mds, mc->reg, r->pool);
+        ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "html table");
+        md_status_get_json(&jstatus, mds, mc->reg, mc->ocsp, r->pool);
+        ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "got JSON status");
         apr_brigade_puts(ctx.bb, NULL, NULL, 
-                         "<hr>\n<h2>Managed Domains</h2>\n<table class='md_status'><thead><tr>\n");
+                         "<hr>\n<h3>Managed Certificates</h3>\n<table class='md_status'><thead><tr>\n");
         for (i = 0; i < (int)(sizeof(status_infos)/sizeof(status_infos[0])); ++i) {
-            apr_brigade_puts(ctx.bb, NULL, NULL, "<th>");
-            apr_brigade_puts(ctx.bb, NULL, NULL, status_infos[i].label);
-            apr_brigade_puts(ctx.bb, NULL, NULL, "</th>");
+            si_add_header(&ctx, &status_infos[i]);
         }
         apr_brigade_puts(ctx.bb, NULL, NULL, "</tr>\n</thead><tbody>");
         md_json_itera(add_md_row, &ctx, jstatus, MD_KEY_MDS, NULL);
@@ -483,12 +568,101 @@ int md_status_hook(request_rec *r, int flags)
 
     ap_pass_brigade(r->output_filters, ctx.bb);
     apr_brigade_cleanup(ctx.bb);
+    ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "server-status for managed domains, end");
+    
+    return OK;
+}
+
+static void si_val_ocsp_activity(status_ctx *ctx, md_json_t *mdj, const status_info *info)
+{
+    apr_time_t t;
+    
+    (void)info;
+    t = md_json_get_time(mdj,  MD_KEY_RENEW_AT, NULL);
+    print_time(ctx->bb, "Refresh", t);
+    print_job_summary(ctx->bb, mdj, MD_KEY_RENEWAL, ": ");
+}
+
+static const status_info ocsp_status_infos[] = {
+    { "Domain", MD_KEY_DOMAIN, NULL },
+    { "Certificate ID", MD_KEY_ID, NULL },
+    { "OCSP Status", MD_KEY_STATUS, NULL },
+    { "Stapling Valid", MD_KEY_VALID, si_val_valid_time },
+    { "Responder", MD_KEY_URL, si_val_url },
+    { "Activity",  MD_KEY_NOTIFIED, si_val_ocsp_activity },
+};
+
+static int add_ocsp_row(void *baton, apr_size_t index, md_json_t *mdj)
+{
+    status_ctx *ctx = baton;
+    int i;
+    
+    apr_brigade_printf(ctx->bb, NULL, NULL, "<tr class=\"%s\">", (index % 2)? "odd" : "even");
+    for (i = 0; i < (int)(sizeof(ocsp_status_infos)/sizeof(ocsp_status_infos[0])); ++i) {
+        apr_brigade_puts(ctx->bb, NULL, NULL, "<td>");
+        add_status_cell(ctx, mdj, &ocsp_status_infos[i]);
+        apr_brigade_puts(ctx->bb, NULL, NULL, "</td>");
+    }
+    apr_brigade_puts(ctx->bb, NULL, NULL, "</tr>");
+    return 1;
+}
+
+int md_ocsp_status_hook(request_rec *r, int flags)
+{
+    const md_srv_conf_t *sc;
+    const md_mod_conf_t *mc;
+    int i, html;
+    status_ctx ctx;
+    md_json_t *jstatus, *jstock;
+    
+    ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "server-status for ocsp stapling, start");
+    sc = ap_get_module_config(r->server->module_config, &md_module);
+    if (!sc) return DECLINED;
+    mc = sc->mc;
+    if (!mc || !mc->server_status_enabled) return DECLINED;
+
+    html = !(flags & AP_STATUS_SHORT);
+    ctx.p = r->pool;
+    ctx.mc = mc;
+    ctx.bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
+    ctx.separator = " ";
+
+    if (!html) {
+        apr_brigade_puts(ctx.bb, NULL, NULL, "Managed Staplings: ");
+        if (md_ocsp_count(mc->ocsp) > 0) {
+            md_ocsp_get_summary(&jstock, mc->ocsp, r->pool);
+            apr_brigade_printf(ctx.bb, NULL, NULL, "total=%d, good=%d revoked=%d unknown=%d",
+                                (int)md_json_getl(jstock, MD_KEY_TOTAL, NULL), 
+                                (int)md_json_getl(jstock, MD_KEY_GOOD, NULL), 
+                                (int)md_json_getl(jstock, MD_KEY_REVOKED, NULL), 
+                                (int)md_json_getl(jstock, MD_KEY_UNKNOWN, NULL));
+        } 
+        else {
+            apr_brigade_puts(ctx.bb, NULL, NULL, "[]"); 
+        }
+        apr_brigade_puts(ctx.bb, NULL, NULL, "\n"); 
+    }
+    else if (md_ocsp_count(mc->ocsp) > 0) {
+        md_ocsp_get_status_all(&jstatus, mc->ocsp, r->pool);
+        apr_brigade_puts(ctx.bb, NULL, NULL, 
+                         "<hr>\n<h3>Managed Staplings</h3>\n<table class='md_ocsp_status'><thead><tr>\n");
+        for (i = 0; i < (int)(sizeof(ocsp_status_infos)/sizeof(ocsp_status_infos[0])); ++i) {
+            si_add_header(&ctx, &ocsp_status_infos[i]);
+        }
+        apr_brigade_puts(ctx.bb, NULL, NULL, "</tr>\n</thead><tbody>");
+        md_json_itera(add_ocsp_row, &ctx, jstatus, MD_KEY_OCSPS, NULL);
+        apr_brigade_puts(ctx.bb, NULL, NULL, "</td></tr>\n</tbody>\n</table>\n");
+    }
+
+    ap_pass_brigade(r->output_filters, ctx.bb);
+    apr_brigade_cleanup(ctx.bb);
+    ap_log_rerror(APLOG_MARK, APLOG_TRACE1, 0, r, "server-status for ocsp stapling, end");
     
     return OK;
 }
 
 /**************************************************************************************************/
-/* Status handler */
+/* Status handlers */
 
 int md_status_handler(request_rec *r)
 {
@@ -523,12 +697,12 @@ int md_status_handler(request_rec *r)
     }
     
     if (md) {
-        md_status_get_md_json(&jstatus, md, mc->reg, r->pool);
+        md_status_get_md_json(&jstatus, md, mc->reg, mc->ocsp, r->pool);
     }
     else {
         mds = apr_array_copy(r->pool, mc->mds);
         qsort(mds->elts, (size_t)mds->nelts, sizeof(md_t *), md_name_cmp);
-        md_status_get_json(&jstatus, mds, mc->reg, r->pool);
+        md_status_get_json(&jstatus, mds, mc->reg, mc->ocsp, r->pool);
     }
 
     if (jstatus) {
@@ -542,3 +716,4 @@ int md_status_handler(request_rec *r)
     }
     return DECLINED;
 }
+
