@@ -57,6 +57,11 @@ POSSIBILITY OF SUCH DAMAGE.
 #include "apr_tables.h"
 #include "pcre.h"
 
+/* PCRE_DUPNAMES is only present since version 6.7 of PCRE */
+#ifndef PCRE_DUPNAMES
+#error PCRE Version 6.7 or later required!
+#else
+
 #define APR_WANT_STRFUNC
 #include "apr_want.h"
 
@@ -120,8 +125,7 @@ AP_DECLARE(void) ap_regfree(ap_regex_t *preg)
  *            Compile a regular expression       *
  *************************************************/
 
-static int default_cflags = AP_REG_DOTALL |
-                            AP_REG_DOLLAR_ENDONLY;
+static int default_cflags = AP_REG_DEFAULT;
 
 AP_DECLARE(int) ap_regcomp_get_default_cflags(void)
 {
@@ -169,7 +173,9 @@ AP_DECLARE(int) ap_regcomp(ap_regex_t * preg, const char *pattern, int cflags)
     int errcode = 0;
     int options = PCRE_DUPNAMES;
 
-    cflags |= default_cflags;
+    if ((cflags & AP_REG_NO_DEFAULT) == 0)
+        cflags |= default_cflags;
+
     if ((cflags & AP_REG_ICASE) != 0)
         options |= PCRE_CASELESS;
     if ((cflags & AP_REG_NEWLINE) != 0)
@@ -341,4 +347,6 @@ AP_DECLARE(int) ap_regname(const ap_regex_t *preg,
     return namecount;
 }
 
+#endif /* PCRE_DUPNAMES defined */
+
 /* End of pcreposix.c */
